package GUT;

import checkers.types.AnnotatedTypeMirror;
import checkers.types.AnnotatedTypeMirror.AnnotatedArrayType;
import checkers.types.AnnotatedTypeMirror.AnnotatedDeclaredType;
import checkers.types.AnnotatedTypeMirror.AnnotatedTypeVariable;
import checkers.types.AnnotatedTypeMirror.AnnotatedWildcardType;
import checkers.util.Pair;
import checkers.util.TypesUtils;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.List;
import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Element;
import javax.lang.model.element.TypeElement;
import javax.lang.model.element.TypeParameterElement;
import javax.lang.model.type.DeclaredType;
import javax.lang.model.type.TypeKind;


/**
 * Utility class that contains Universe type helper methods.
 *
 * @author wmdietl
 */
public class GUTQualifierUtils {

    public static void setGUTChecker(GUTChecker c) {
        checker = c;
    }
    private static GUTChecker checker;

    /**
     * Combine two Universe modifiers.
     *
     * @param recv the receiver Universe modifier.
     * @param decl the declared Universe modifier.
     * @return the combined Universe modifier.
     */
    private static AnnotationMirror combineModifierWithModifier(AnnotationMirror recv, AnnotationMirror decl) {
        assert recv != null;
        assert decl != null;

        if (recv == checker.SELF) {
            return decl;
        } else if (recv == checker.PEER && decl == checker.PEER) {
            return checker.PEER;
        } else if (recv == checker.REP && decl == checker.PEER) {
            return checker.REP;
        } else if (decl == checker.ANY) {
            return checker.ANY;
        } else {
            return checker.LOST;
        }
    }

    // Used to limit the recursion in upper bounds
    private static boolean isTypeVarExtends = false;

    /**
     * Combine a Universe modifier with a type.
     *
     * @param recv the receiver Universe modifier.
     * @param decl the declared type.
     * @return the combined type.
     *
     * TODO: unsupported TypeKinds create assertion violation.
     */
    public static AnnotatedTypeMirror combineModifierWithType(AnnotationMirror recv,
            AnnotatedTypeMirror decl) {
        assert recv != null;
        assert decl != null;

        if (decl.getKind().isPrimitive()) {
        	// Make sure that primitive types have type bottom
        	decl.clearAnnotations();
        	decl.addAnnotation(checker.BOTTOM);
        	return decl;
        } else if (decl.getKind() == TypeKind.TYPEVAR ) {
        	if (!isTypeVarExtends) {
        		isTypeVarExtends=true;

                AnnotatedTypeVariable atv = (AnnotatedTypeVariable) decl.getCopy(true);
                Map<AnnotatedTypeMirror, AnnotatedTypeMirror> mapping = new HashMap<AnnotatedTypeMirror, AnnotatedTypeMirror>();

                AnnotatedTypeMirror resUpper = combineModifierWithType(recv, atv.getUpperBound());
                mapping.put(atv.getUpperBound(), resUpper);

                AnnotatedTypeMirror result = atv.substitute(mapping);
                isTypeVarExtends=false;
                return result;
        	}
        	return decl;
        } else if (decl.getKind() == TypeKind.DECLARED) {
            // Create a copy
            AnnotatedDeclaredType declaredType = (AnnotatedDeclaredType) decl.getCopy(true);
            
            if (TypesUtils.isString(declaredType.getUnderlyingType())) {
                declaredType.clearAnnotations();
                declaredType.addAnnotation(checker.ANY);
                return declaredType;
            }
            
            Map<AnnotatedTypeMirror, AnnotatedTypeMirror> mapping = new HashMap<AnnotatedTypeMirror, AnnotatedTypeMirror>();

            // Get the combined main modifier
            AnnotationMirror mainModifier = getUniverseModifier(declaredType);
            AnnotationMirror combinedMainModifier = combineModifierWithModifier(recv, mainModifier);

            // Get the combined type arguments
            for (AnnotatedTypeMirror typeArgument : declaredType.getTypeArguments()) {
                AnnotatedTypeMirror combinedTypeArgument = combineModifierWithType(recv, typeArgument);
                mapping.put(typeArgument, combinedTypeArgument);
            }

            // Construct result type
            AnnotatedTypeMirror result = declaredType.substitute(mapping);
            result.clearAnnotations();
            result.addAnnotation(combinedMainModifier);

            return result;
        } else if (decl.getKind() == TypeKind.ARRAY) {
            // Create a copy
            AnnotatedArrayType  arrayType = (AnnotatedArrayType) decl.getCopy(true);
            // Get the combined main modifier
            AnnotationMirror mainModifier = getUniverseModifier(arrayType);
            AnnotationMirror combinedMainModifier = combineModifierWithModifier(recv, mainModifier);

            // Construct result type
            arrayType.clearAnnotations();
            arrayType.addAnnotation(combinedMainModifier);

            // TODO: combine element type
            AnnotatedTypeMirror elemType = arrayType.getComponentType();
   			AnnotatedTypeMirror combinedElemType = combineModifierWithType(recv, elemType);

   			arrayType.setComponentType(combinedElemType);
            return arrayType;
        } else if (decl.getKind() == TypeKind.WILDCARD) {
            // Create a copy
            AnnotatedWildcardType wildType = (AnnotatedWildcardType) decl.getCopy(true);
            Map<AnnotatedTypeMirror, AnnotatedTypeMirror> mapping = new HashMap<AnnotatedTypeMirror, AnnotatedTypeMirror>();

            // There is no main modifier for a wildcard

            AnnotatedTypeMirror combined;
            if (wildType.getExtendsBound() != null) {
            	combined = combineModifierWithType(recv, wildType.getExtendsBound());
            	mapping.put(wildType.getExtendsBound(), combined);
            }
            if (wildType.getSuperBound() != null) {
            	combined = combineModifierWithType(recv, wildType.getSuperBound());
            	mapping.put(wildType.getSuperBound(), combined);
            }

            // Construct result type
            AnnotatedTypeMirror result = wildType.substitute(mapping);
            return result;
        } else {
        	System.err.println("Unknown result.getKind(): " + decl.getKind());
            assert false;
            return null;
            // TODO throw new UnsupportedTypeException(type.getKind().toString());
        }
    }

    /**
     * Combine two types.
     *
     * @param recv the receiver type.
     * @param decl the declared type.
     * @return the combined type.
     */
    public static AnnotatedTypeMirror combineTypeWithType(AnnotatedTypeMirror recv, AnnotatedTypeMirror decl) {
        assert recv != null;
        assert decl != null;

        // System.out.println("combTT recv: " + recv);
        // System.out.println("combTT decl: " + decl);

        if (recv.getKind() == TypeKind.TYPEVAR) {
        	recv = ((AnnotatedTypeVariable)recv).getUpperBound();
        	// System.out.println("combTT using upper bound: " + recv);
        }

        AnnotationMirror firstMainModifier = getUniverseModifier(recv);
        AnnotatedTypeMirror combinedType = combineModifierWithType(firstMainModifier, decl);

        combinedType = substituteTVars(recv, combinedType);

        // System.out.println("combTT comb: " + combinedType);
        return combinedType;
    }

    private static AnnotatedTypeMirror substituteTVars(AnnotatedTypeMirror lhs, AnnotatedTypeMirror rhs) {
        if (rhs.getKind() == TypeKind.TYPEVAR) {
        	// method type variables will never be found and no special case needed; maybe as optimization.
            AnnotatedTypeVariable rhsTypeVar = (AnnotatedTypeVariable) rhs.getCopy(true);

        	if (lhs.getKind() == TypeKind.DECLARED) {
        		rhs = getTypeVariableSubstitution((AnnotatedDeclaredType) lhs, rhsTypeVar);
        	}
        	// else TODO: the receiver might be another type variable... should we do something?
        } else if (rhs.getKind() == TypeKind.DECLARED) {
            // Create a copy
            AnnotatedDeclaredType declaredType = (AnnotatedDeclaredType) rhs.getCopy(true);
            Map<AnnotatedTypeMirror, AnnotatedTypeMirror> mapping = new HashMap<AnnotatedTypeMirror, AnnotatedTypeMirror>();

            for (AnnotatedTypeMirror typeArgument : declaredType.getTypeArguments()) {
                AnnotatedTypeMirror substTypeArgument = substituteTVars(lhs, typeArgument);
                mapping.put(typeArgument, substTypeArgument);
            }

            // Construct result type
            rhs = declaredType.substitute(mapping);
        } else if (rhs.getKind() == TypeKind.WILDCARD) {
            AnnotatedWildcardType wildType = (AnnotatedWildcardType) rhs.getCopy(true);
            Map<AnnotatedTypeMirror, AnnotatedTypeMirror> mapping = new HashMap<AnnotatedTypeMirror, AnnotatedTypeMirror>();

            AnnotatedTypeMirror substBound;

            if (wildType.getExtendsBound()!=null) {
            	substBound = substituteTVars(lhs, wildType.getExtendsBound());
            	mapping.put(wildType.getExtendsBound(), substBound);
            }
            if (wildType.getSuperBound()!=null) {
            	substBound = substituteTVars(lhs, wildType.getSuperBound());
            	mapping.put(wildType.getSuperBound(), substBound);
            }

            rhs = wildType.substitute(mapping);
        } else if (rhs.getKind() == TypeKind.ARRAY) {
            // Create a copy
            AnnotatedArrayType arrayType = (AnnotatedArrayType) rhs.getCopy(true);

            Map<AnnotatedTypeMirror, AnnotatedTypeMirror> mapping = new HashMap<AnnotatedTypeMirror, AnnotatedTypeMirror>();

            AnnotatedTypeMirror compType = arrayType.getComponentType();
            AnnotatedTypeMirror substTypeArgument = substituteTVars(lhs, compType);
            mapping.put(compType, substTypeArgument);

            // Construct result type
            rhs = arrayType.substitute(mapping);
        } else if (rhs.getKind().isPrimitive()) {
        	// nothing to do for primitive types
        } else {
        	System.out.println("GUTQualifierUtils::substituteTVars: What should be done with: " + rhs + " of kind: " + rhs.getKind());
        }

        return rhs;
    }

    private static AnnotatedTypeMirror getTypeVariableSubstitution(AnnotatedDeclaredType type, AnnotatedTypeVariable var) {
    	// System.out.println("Looking for " + var + " in type " + type);

        Pair<AnnotatedDeclaredType, Integer> res = findDeclType(type, var);

    	if(res==null) {
    		return var;
    	}

    	AnnotatedDeclaredType decltype = res.first;
    	int foundindex = res.second;

		if (decltype.isParameterized()) {
    		List<AnnotatedTypeMirror> tas = decltype.getTypeArguments();
			// assert foundindex < tas.size()) {
    		// CAREFUL: return a copy, as we want to modify the type later.
			return tas.get(foundindex).getCopy(true);
		} else {
			// TODO: use upper bound instead of var
			// type.getTypeArguments()
			// System.out.println("Raw Type: " + decltype);
			if (!var.getUpperBound().isAnnotated()) {
				// TODO: hmm, seems to be needed :-(
				var.getUpperBound().addAnnotation(checker.PEER);
			}
			return var.getUpperBound();
		}
    }

    private static Pair<AnnotatedDeclaredType, Integer> findDeclType(AnnotatedDeclaredType type, AnnotatedTypeVariable var) {
    	// System.out.println("Finding " + var + " in type " + type);

    	Element varelem = var.getUnderlyingType().asElement();

    	DeclaredType dtype = type.getUnderlyingType();
    	TypeElement el = (TypeElement) dtype.asElement();
    	List<? extends TypeParameterElement> tparams = el.getTypeParameters();
    	int foundindex = 0;

    	for (TypeParameterElement tparam: tparams) {
    		if (tparam.equals(varelem) ||
    				//TODO: comparing by name!!!???
    				// Sometimes "E" and "E extends Object" are compared, which do not match by "equals".
    				tparam.getSimpleName().equals(varelem.getSimpleName())) {
    			// we found the right index!
    			break;
    		}
    		++foundindex;
    	}

    	if (foundindex>=tparams.size()) {
    		// didn't find the desired type :-(
    		for(AnnotatedDeclaredType sup : type.directSuperTypes()) {
    		    Pair<AnnotatedDeclaredType, Integer> res = findDeclType(sup, var);

    	    	if(res!=null) {
    	    		return res;
    	    	}
    		}
    		// we reach this point if the variable wasn't found in any recursive call.
    		return null;
    	}

    	return Pair.of(type, foundindex);
    }


    /**
     * Returns the first Universe modifier on the annotated type.
     *
     * @param type an annotated type.
     * @return the first Universe modifier or null if there's no Universe modifier.
     */
    public static AnnotationMirror getUniverseModifier(AnnotatedTypeMirror type) {
        assert type != null;

        if (!type.isAnnotated()) {
        	// Hmm, why do we get here with an unannotated type?
        	// Seems to happen for arrays?
        	if (GUTChecker.isAnyDefault(type) ) {
    			return checker.ANY;
    		} else {
    			return checker.PEER;
    		}
        }

        if (type.hasAnnotation(checker.PEER)) {
            return checker.PEER;
        }
        if (type.hasAnnotation(checker.REP)) {
            return checker.REP;
        }
        if (type.hasAnnotation(checker.ANY)) {
            return checker.ANY;
        }
        if (type.hasAnnotation(checker.SELF)) {
            return checker.SELF;
        }

        return checker.LOST;

        // TODO: for some reason the simpler statement below does not work!
        // It gives a different AnnotationMirror for different invocations.
        // Defaulting ensures that there is always one Universe annotation.
        // return type.getAnnotations().iterator().next();
    }


    /**
     * Checks whether type has multiple Universe modifiers on the top level.
     *
     * @param type a type.
     * @return true if type has multiple Universe modifiers, false otherwise.
     */
    public static boolean hasMultipleModifiers(AnnotatedTypeMirror type) {
        assert type != null;

        return type.getAnnotations().size() > 1;
    }
}
