package GUT;

import java.util.List;
import java.util.Set;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Modifier;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.DeclaredType;

import checkers.basetype.BaseTypeVisitor;
import checkers.source.Result;
import checkers.types.AnnotatedTypeMirror;
import checkers.types.AnnotatedTypeMirror.AnnotatedArrayType;
import checkers.types.AnnotatedTypeMirror.AnnotatedDeclaredType;
import checkers.types.AnnotatedTypeMirror.AnnotatedExecutableType;
import checkers.types.AnnotatedTypeMirror.AnnotatedPrimitiveType;
import checkers.types.AnnotatedTypeMirror.AnnotatedTypeVariable;
import checkers.types.AnnotatedTypes;
import checkers.util.ElementUtils;
import checkers.util.Pair;
import checkers.util.TreeUtils;

import com.sun.source.tree.AssignmentTree;
import com.sun.source.tree.CompilationUnitTree;
import com.sun.source.tree.ExpressionTree;
import com.sun.source.tree.MethodInvocationTree;
import com.sun.source.tree.MethodTree;
import com.sun.source.tree.ModifiersTree;
import com.sun.source.tree.NewClassTree;
import com.sun.source.tree.ParameterizedTypeTree;
import com.sun.source.tree.Tree;
import com.sun.source.tree.TypeCastTree;
import com.sun.source.tree.VariableTree;
import com.sun.source.tree.Tree.Kind;

/**
 * Type visitor to enforce the GUT type rules.
 *
 * @author wmdietl
 */
public class GUTVisitor extends BaseTypeVisitor<GUTChecker> {
	public GUTVisitor(GUTChecker checker, CompilationUnitTree root) {
		super(checker, root);
	}

	/**
	 * The type validator to ensure correct usage of ownership modifiers.
	 */
	@Override
	protected TypeValidator createTypeValidator() {
		return new GUTTypeValidator();
	}

	/**
	 * GUT does not use receiver annotations, forbid them.
	 */
	@Override
	public Void visitMethod(MethodTree node, Void p) {
		// System.out.println("MethodTree: " + node);

		if (!node.getReceiverAnnotations().isEmpty()) {
			checker.report(Result.failure("uts.receiver.annotations.forbidden"), node);
		}

		return super.visitMethod(node, p);
	}

	/**
	 * Ignore method receiver annotations.
	 */
	@Override
	protected boolean checkMethodInvocability(AnnotatedExecutableType method,
			MethodInvocationTree node) {
		return true;
	}

	/**
	 * Ignore constructor receiver annotations.
	 */
	@Override
	protected boolean checkConstructorInvocation(AnnotatedDeclaredType dt,
			AnnotatedExecutableType constructor, Tree src) {
		return true;
	}

	/**
	 * Validate a new object creation.
	 *
	 * @param node
	 *            the new object creation.
	 * @param p
	 *            not used.
	 */
	@Override
	public Void visitNewClass(NewClassTree node, Void p) {
		assert node != null;

		Pair<AnnotatedExecutableType, List<AnnotatedTypeMirror>> fromUse = atypeFactory.constructorFromUse(node);
		AnnotatedExecutableType constructor = fromUse.first;

		// Check for @Lost in combined parameter types.
		for (AnnotatedTypeMirror parameterType : constructor.getParameterTypes()) {
			if (AnnotatedTypes.containsModifier(parameterType, checker.LOST)) {
				checker.report(Result.failure("uts.lost.parameter"), node);
			}
		}

		// Check for @Peer or @Rep as top-level modifier.
		AnnotatedTypeMirror type = atypeFactory.getAnnotatedType(node);
		if (!(type.hasAnnotation(checker.PEER) || type.hasAnnotation(checker.REP))) {
			// System.out.println("new got: " + type.toString());
			checker.report(Result.failure("uts.new.ownership"), node);
		}

		// TODO: forbid rep in static context
		// TODO: test this. fields, methods, local vars, ... Checker.isValid?
		if (ElementUtils.isStatic(TreeUtils.elementFromUse(node))
				&& type.hasAnnotation(checker.REP)) {
			checker.report(Result.failure("uts.static.rep.forbidden"), node);
		}

		return super.visitNewClass(node, p);
	}

	/**
	 * Validate a method invocation.
	 *
	 * @param node
	 *            the method invocation.
	 * @param p
	 *            not used.
	 */
	@Override
	public Void visitMethodInvocation(MethodInvocationTree node, Void p) {
		assert node != null;

		AnnotatedExecutableType method = atypeFactory.methodFromUse(node).first;

		// Check for @Lost in combined parameter types.
		for (AnnotatedTypeMirror parameterType : method.getParameterTypes()) {
			if (AnnotatedTypes.containsModifier(parameterType, checker.LOST)) {
				checker.report(Result.failure("uts.lost.parameter"), node);
			}
		}

        if( (checker).checkOaM) {
        	ExpressionTree recvTree = TreeUtils.getReceiverTree(node.getMethodSelect());
            if (recvTree != null) {
            	AnnotatedTypeMirror recvType = atypeFactory.getAnnotatedType(recvTree);

            	if (recvType != null) {
            		if (recvType.hasAnnotation(checker.LOST) ||
            				recvType.hasAnnotation(checker.ANY)) {
            			ExecutableElement exelem = TreeUtils.elementFromUse(node);
            			java.util.List<? extends AnnotationMirror> anns = exelem.getAnnotationMirrors();

            			boolean hasPure = false;
                        for (AnnotationMirror an : anns) {
                          if (isPure(an.getAnnotationType())) {
                            hasPure = true;
                          }
                        }
                        if (anns.isEmpty() || !hasPure) {
                        	checker.report(Result.failure("oam.call.forbidden"), node);
                        }
            		}
            	}
            }
        }

		// TODO purity

		return super.visitMethodInvocation(node, p);
	}

	private boolean isPure(DeclaredType anno) {
	    // TODO: Is this the simplest way to do this?
        return anno.toString().equals(GUT.quals.Pure.class.getName())
                || anno.toString().equals(org.jmlspecs.annotation.Pure.class.getName());
    }

    /**
	 * Validate an assignment.
	 *
	 * @param node
	 *            the assignment.
	 * @param p
	 *            not used.
	 */
	@Override
	public Void visitAssignment(AssignmentTree node, Void p) {
		assert node != null;

		// Check for @Lost in left hand side of assignment.
		AnnotatedTypeMirror type = atypeFactory.getAnnotatedType(node.getVariable());
        if (AnnotatedTypes.containsModifier(type, checker.LOST)) {
			checker.report(Result.failure("uts.lost.lhs"), node);
        }

        if( (checker).checkOaM) {
        	ExpressionTree recvTree = TreeUtils.getReceiverTree(node.getVariable());
            if (recvTree != null) {
            	AnnotatedTypeMirror recvType = atypeFactory.getAnnotatedType(recvTree);

            	if (recvType != null) {
            		if (recvType.hasAnnotation(checker.LOST) ||
            				recvType.hasAnnotation(checker.ANY)) {
            			checker.report(Result.failure("oam.assignment.forbidden"), node);
            		}
            	}
            }
        }

        if( (checker).checkStrictPurity && true /*TODO environment pure*/) {
        	checker.report(Result.failure("purity.assignment.forbidden"), node);
        }

        return super.visitAssignment(node, p);
	}

	@Override
	public Void visitTypeCast(TypeCastTree node, Void p) {
		AnnotatedTypeMirror type = atypeFactory
				.getAnnotatedType(node.getType());
		// AnnotatedTypeMirror expr =
		// atypeFactory.getAnnotatedType(node.getExpression());

		if ((AnnotatedTypes.containsModifier(type, checker.LOST)
		        || AnnotatedTypes.containsModifier(type, checker.ANY))
				&& !GUTChecker.isAnyDefault(type)) {
			checker.report(Result.warning("uts.cast.type.warning", type), node);
			// checker.getProcessingEnvironment().getMessager().printMessage(javax.tools.Diagnostic.Kind.WARNING,
			// "Casting to " + type + " is not recommended.");
		}

		// The only part from the super call that we want.
		validateTypeOf(node.getType());

		// calling super would check for cast safety... is there a different way
		// to unset cast:unsafe?
		return null;
		// return super.visitTypeCast(node, p);
	}

	/**
	 * This type validator ensures correct usage of ownership modifiers.
	 */
	private final class GUTTypeValidator extends TypeValidator {

		/**
		 * Ensure that only one ownership modifier is used, that ownership
		 * modifiers are correctly used in static contexts, and check for
		 * explicit use of lost.
		 */
		@Override
		public Void visitDeclared(AnnotatedDeclaredType type, Tree p) {
			if (p.getKind() == Kind.VARIABLE) {
				ModifiersTree mt = ((VariableTree) p).getModifiers();
				if (mt.getFlags().contains(Modifier.STATIC)) {
					if (AnnotatedTypes.containsModifier(type, checker.REP)) {
						checker.report(Result.failure("uts.static.rep.forbidden",
						        type.getAnnotations(), type.toString()), p);
					}
					if (AnnotatedTypes.containsModifier(type, checker.PEER)
							&& (checker).warn_staticpeer) {
						// TODO: I would really like to only give the warning if
						// the modifier was explicit.
						checker.report(Result.warning("uts.static.peer.warning",
						        type.getAnnotations(), type.toString()), p);
					}
				} else {
					// TODO: how do I find if I'm in a static method? selftype
					// is non-null :-(
					// System.out.println("Self: " +
					// GUTVisitor.this.atypeFactory.getSelfType(p));
					// System.out.println("Check on tree: " + p);
				}
			}

			if (!(checker).allowLost
					&& AnnotatedTypes.containsModifier(type, checker.LOST)) {
				checker.report(Result.failure("uts.explicit.lost.forbidden",
				        type.getAnnotations(), type.toString()), p);
			}

			if (GUTQualifierUtils.hasMultipleModifiers(type)) {
				reportError(type, p);
			}
			return super.visitDeclared(type, p);
		}

        @Override
        protected Void visitParameterizedType(AnnotatedDeclaredType type, ParameterizedTypeTree tree) {
            final TypeElement element =
                (TypeElement)type.getUnderlyingType().asElement();

            List<AnnotatedTypeVariable> typevars = atypeFactory.typeVariablesFromUse(type, element);

            for (AnnotatedTypeVariable atv : typevars) {
                if (AnnotatedTypes.containsModifier(atv, checker.LOST)) {
                    checker.report(Result.failure("uts.lost.in.bounds",
                            atv.toString(), type.toString()), tree);
                }
            }
            return super.visitParameterizedType(type, tree);
        }

        /**
		 * Forbid explicit annotations on primitive types.
		 */
		@Override
		public Void visitPrimitive(AnnotatedPrimitiveType type, Tree p) {
			if (type.isAnnotated()) {
				Set<AnnotationMirror> ann = type.getAnnotations();
				if (ann.size() > 1
						|| (ann.size() == 1 && !ann.contains(checker.BOTTOM))) {
					// the implicit default is BOTTOM, which cannot be used
					// explicitly.
					// if there are explicit annotations -> error.
					reportError(type, p);
				}
			}
			return super.visitPrimitive(type, p);
		}

		/**
		 * Each array dimension can only use at most one ownership modifier. The
		 * check of the component type is done by the super method.
		 */
		@Override
		public Void visitArray(AnnotatedArrayType type, Tree p) {
			if (type.getAnnotations().size() > 1) {
				// only one Universe modifier is allowed on the array
				reportError(type, p);
			}
			return super.visitArray(type, p);
		}
	}
}
