package GUTI;

import GUT.*;

import java.util.List;
import java.util.Set;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.DeclaredType;

import checkers.basetype.BaseTypeChecker;
import checkers.inference.InferenceVisitor;
import checkers.source.Result;
import checkers.types.AnnotatedTypeMirror;
import checkers.types.AnnotatedTypeMirror.AnnotatedArrayType;
import checkers.types.AnnotatedTypeMirror.AnnotatedDeclaredType;
import checkers.types.AnnotatedTypeMirror.AnnotatedExecutableType;
import checkers.types.AnnotatedTypeMirror.AnnotatedPrimitiveType;
import checkers.types.AnnotatedTypeMirror.AnnotatedTypeVariable;
import checkers.types.AnnotatedTypes;
import checkers.util.ElementUtils;
import checkers.util.Pair;
import checkers.util.TreeUtils;

import com.sun.source.tree.AssignmentTree;
import com.sun.source.tree.BinaryTree;
import com.sun.source.tree.BlockTree;
import com.sun.source.tree.CompilationUnitTree;
import com.sun.source.tree.ExpressionTree;
import com.sun.source.tree.InstanceOfTree;
import com.sun.source.tree.MethodInvocationTree;
import com.sun.source.tree.MethodTree;
import com.sun.source.tree.NewClassTree;
import com.sun.source.tree.ParameterizedTypeTree;
import com.sun.source.tree.Tree;
import com.sun.source.tree.TypeCastTree;
import com.sun.source.tree.VariableTree;
import com.sun.source.tree.Tree.Kind;
import com.sun.source.util.TreePath;

/**
 * Type visitor to either enforce or infer the GUT type rules.
 *
 * @author wmdietl
 */
public class GUTIVisitor extends InferenceVisitor {

    private final boolean checkOaM;
    private final boolean checkStrictPurity;
    private final boolean allowLost;

    private final boolean warn_staticpeer;

    private final GUTIChecker gutichecker;

    /*
     * We continue to use BaseTypeChecker as first parameter type, because this
     * class is instantiated by both the GUTIChecker and the InferenceChecker.
     */
    public GUTIVisitor(BaseTypeChecker checker, CompilationUnitTree root, GUTIChecker gutichecker, boolean infer) {
        super(checker, root, infer);

        this.gutichecker = gutichecker;

        if (checker instanceof GUTIChecker) {
            GUTIChecker gc = (GUTIChecker) checker;

            checkOaM = gc.checkOaM;
            checkStrictPurity = gc.checkStrictPurity;
            allowLost = gc.allowLost;

            warn_staticpeer = gc.warn_staticpeer;
        } else {
            checkOaM = false;
            checkStrictPurity = false;
            allowLost = false;
            warn_staticpeer = false;

            // TODO: get the booleans from ??? Pass in TTIRun?
        }
    }

    /**
     * The type validator to ensure correct usage of ownership modifiers.
     */
    @Override
    protected GUTITypeValidator createTypeValidator() {
        return new GUTITypeValidator();
    }

    /**
     * GUT does not use receiver annotations, forbid them.
     */
    @Override
    public Void visitMethod(MethodTree node, Void p) {
        // System.out.println("MethodTree: " + node);

        if (!node.getReceiverAnnotations().isEmpty()) {
            checker.report(Result.failure("uts.receiver.annotations.forbidden"), node);
        }

        return super.visitMethod(node, p);
    }

    /**
     * Ignore method receiver annotations.
     */
    @Override
    protected boolean checkMethodInvocability(AnnotatedExecutableType method, MethodInvocationTree node) {
        return true;
    }

    /**
     * Ignore constructor receiver annotations.
     */
    @Override
    protected boolean checkConstructorInvocation(AnnotatedDeclaredType dt,
            AnnotatedExecutableType constructor, Tree src) {
        return true;
    }

    /**
     * Validate a new object creation.
     *
     * @param node
     *            the new object creation.
     * @param p
     *            not used.
     */
    @Override
    public Void visitNewClass(NewClassTree node, Void p) {
        assert node != null;

        Pair<AnnotatedExecutableType, List<AnnotatedTypeMirror>> fromUse = atypeFactory.constructorFromUse(node);
        AnnotatedExecutableType constructor = fromUse.first;

        // Check for @Lost in combined parameter types.
        for (AnnotatedTypeMirror parameterType : constructor.getParameterTypes()) {
            doesNotContain(parameterType, gutichecker.LOST, "uts.lost.parameter", node);
        }

        // Check for @Peer or @Rep as top-level modifier.
        AnnotatedTypeMirror type = atypeFactory.getAnnotatedType(node);
        mainIsNoneOf(type, new AnnotationMirror[] {gutichecker.LOST, gutichecker.ANY}, "uts.new.ownership", node);

        // Forbid rep in static context
        if (isContextStatic(atypeFactory.getPath(node))) {
            doesNotContain(type, gutichecker.REP, "uts.static.rep.forbidden", node);
        }

        return super.visitNewClass(node, p);
    }

    private static boolean isContextStatic(TreePath path) {
        // TODO: test this. fields, methods, local vars, ... Checker.isValid?
        MethodTree encm = TreeUtils.enclosingMethod(path);
        if (encm!=null) {
            // we are within a method
            return ElementUtils.isStatic(TreeUtils.elementFromDeclaration(encm));
        } else {
            VariableTree encv = TreeUtils.enclosingVariable(path);
            if (encv!=null) {
                // we are within a field initializer
                return ElementUtils.isStatic(TreeUtils.elementFromDeclaration(encv));
            } else {
                // is it a static initializer?
                BlockTree encb = TreeUtils.enclosingTopLevelBlock(path);
                if (encb!=null) {
                    return encb.isStatic();
                } else {
                    // We are somewhere in a class declaration already, e.g. the extends clause
                    // System.out.println("GUTIVisitor::isContextStatic: lost in the tree, help! Path leaf: " + path.getLeaf());
                    return false;
                }
            }
        }
    }

    /**
     * Validate a method invocation.
     *
     * @param node
     *            the method invocation.
     * @param p
     *            not used.
     */
    @Override
    public Void visitMethodInvocation(MethodInvocationTree node, Void p) {
        assert node != null;

        AnnotatedExecutableType method = atypeFactory.methodFromUse(node).first;

        // Check for @Lost in combined parameter types.
        for (AnnotatedTypeMirror parameterType : method.getParameterTypes()) {
            doesNotContain(parameterType, gutichecker.LOST, "uts.lost.parameter", node);
        }

        if (checkOaM) {
            ExpressionTree recvTree = TreeUtils.getReceiverTree(node.getMethodSelect());
            if (recvTree != null) {
                AnnotatedTypeMirror recvType = atypeFactory.getAnnotatedType(recvTree);

                if (recvType != null) {
                    ExecutableElement exelem = TreeUtils.elementFromUse(node);
                    if (!hasPure(exelem)) {
                        mainIsNoneOf(recvType, new AnnotationMirror[] {gutichecker.LOST, gutichecker.ANY},
                                "oam.call.forbidden", node);
                    }
                }
            }
        }

        // TODO purity

        return super.visitMethodInvocation(node, p);
    }

    private static boolean hasPure(ExecutableElement meth) {
        boolean hasPure = false;
        java.util.List<? extends AnnotationMirror> anns = meth.getAnnotationMirrors();
        for (AnnotationMirror an : anns) {
            if (isPure(an.getAnnotationType())) {
                hasPure = true;
                break;
            }
        }
        return hasPure;
    }

    private static boolean isPure(DeclaredType anno) {
        // TODO: Is this the simplest way to do this?
        return anno.toString().equals(GUT.quals.Pure.class.getName())
                || anno.toString().equals(org.jmlspecs.annotation.Pure.class.getName());
    }

    /**
     * Validate an assignment.
     *
     * @param node
     *            the assignment.
     * @param p
     *            not used.
     */
    @Override
    public Void visitAssignment(AssignmentTree node, Void p) {
        assert node != null;

        // Check for @Lost in left hand side of assignment.
        AnnotatedTypeMirror type = atypeFactory.getAnnotatedType(node.getVariable());
        doesNotContain(type, gutichecker.LOST, "uts.lost.lhs", node);

        if (checkOaM) {
            ExpressionTree recvTree = TreeUtils.getReceiverTree(node.getVariable());
            if (recvTree != null) {
                AnnotatedTypeMirror recvType = atypeFactory.getAnnotatedType(recvTree);

                if (recvType != null) {
                    // TODO: do we need to treat "this" and "super" specially?
                    mainIsNoneOf(recvType, new AnnotationMirror[] {gutichecker.LOST, gutichecker.ANY},
                            "oam.assignment.forbidden", node);
                }
            }
        }

        if (checkStrictPurity && true /* TODO environment pure */) {
            checker.report(Result.failure("purity.assignment.forbidden"), node);
        }

        return super.visitAssignment(node, p);
    }

    @Override
    public Void visitTypeCast(TypeCastTree node, Void p) {
        AnnotatedTypeMirror castty = atypeFactory.getAnnotatedType(node.getType());
        AnnotatedTypeMirror exprty = atypeFactory.getAnnotatedType(node.getExpression());

        /* TODO: this warning will only be useful in type checking mode.
         * Would it be helpful to somehow support such warnings for type inference, e.g.
         * by reducing weighting?
         */
        if ((AnnotatedTypes.containsModifier(castty, gutichecker.LOST) ||
                AnnotatedTypes.containsModifier(castty, gutichecker.ANY))
                && !GUTIChecker.isAnyDefault(castty)) {
            checker.report(Result.warning("uts.cast.type.warning", castty), node);
            // checker.getProcessingEnvironment().getMessager().printMessage(javax.tools.Diagnostic.Kind.WARNING,
            // "Casting to " + type + " is not recommended.");
        }

        areComparable(castty, exprty, "uts.cast.type.error", node);

        // The only part from the super call that we want.
        validateTypeOf(node.getType());

        // calling super would check for cast safety... is there a different way
        // to unset cast:unsafe?
        return null;
        // return super.visitTypeCast(node, p);
    }

    @Override
    public Void visitInstanceOf(InstanceOfTree node, Void p) {
        AnnotatedTypeMirror ioty = atypeFactory.getAnnotatedType(node.getType());
        AnnotatedTypeMirror exprty = atypeFactory.getAnnotatedType(node.getExpression());

        // TODO: See comment at casts above.
        if ((AnnotatedTypes.containsModifier(ioty, gutichecker.LOST) ||
                AnnotatedTypes.containsModifier(ioty, gutichecker.ANY))
                && !GUTIChecker.isAnyDefault(ioty)) {
            checker.report(Result.warning("uts.instanceof.type.warning", ioty), node);
            // checker.getProcessingEnvironment().getMessager().printMessage(javax.tools.Diagnostic.Kind.WARNING,
            // "Casting to " + type + " is not recommended.");
        }

        areComparable(ioty, exprty, "uts.instanceof.type.error", node);

        return super.visitInstanceOf(node, p);
    }

    /* TODO: what do we want to enforce for strings? They don't really matter anyway.
    @Override
    public Void visitBinary(BinaryTree node, Void p) {
        if (node.getKind() == Kind.PLUS || node.getKind() == Kind.PLUS_ASSIGNMENT) {
            // TODO: only for Strings, but adding it for primitives shouldn't be a problem

            AnnotatedTypeMirror ltype = atypeFactory.getAnnotatedType(node.getLeftOperand());
            AnnotatedTypeMirror rtype = atypeFactory.getAnnotatedType(node.getRightOperand());

            areEqual(ltype, rtype, "uts.append.type.error", node);
        }
        return super.visitBinary(node, p);
    }
    */

    /* TODO Do I really need this?
    @Override
    protected void checkTypeArguments(Tree t,
            java.util.List<? extends AnnotatedTypeVariable> typevars,
            java.util.List<? extends AnnotatedTypeMirror> typeargs,
            java.util.List<? extends Tree> typeargTrees) {
        super.checkTypeArguments(t, typevars, typeargs, typeargTrees);
    }
    */

    /**
     * This type validator ensures correct usage of ownership modifiers.
     */
    private final class GUTITypeValidator extends TypeValidator {

        /**
         * Ensure that only one ownership modifier is used, that ownership
         * modifiers are correctly used in static contexts, and check for
         * explicit use of lost.
         */
        @Override
        public Void visitDeclared(AnnotatedDeclaredType type, Tree p) {

            // ModifiersTree mt = ((VariableTree) p).getModifiers();
            // if (mt.getFlags().contains(Modifier.STATIC)) {
            if (GUTIVisitor.isContextStatic(atypeFactory.getPath(p))) {
                doesNotContain(type, gutichecker.REP, "uts.static.rep.forbidden", p);

                if (warn_staticpeer) {
                    // TODO: I would really like to only give the warning if
                    // the modifier was explicit.
                    // TODO: Only want a warning, not an error.
                    doesNotContain(type, gutichecker.PEER, "uts.static.peer.warning", p);
                }
            }

            if (!allowLost) {
                doesNotContain(type, gutichecker.LOST, "uts.explicit.lost.forbidden", p);
            }

            if (GUTQualifierUtils.hasMultipleModifiers(type)) {
                reportError(type, p);
            }
            // System.out.println("super.visitDeclared for: " + type);
            return super.visitDeclared(type, p);
        }

        @Override
        protected Void visitParameterizedType(AnnotatedDeclaredType type, ParameterizedTypeTree tree) {
            final TypeElement element = (TypeElement) type.getUnderlyingType().asElement();

            List<AnnotatedTypeVariable> typevars = atypeFactory.typeVariablesFromUse(type, element);
            for (AnnotatedTypeVariable atv : typevars) {
                doesNotContain(atv, gutichecker.LOST, "uts.lost.in.bounds", tree);
            }

            return super.visitParameterizedType(type, tree);
        }

        /**
         * Forbid explicit annotations on primitive types.
         */
        @Override
        public Void visitPrimitive(AnnotatedPrimitiveType type, Tree p) {
            if (type.isAnnotated()) {
                Set<AnnotationMirror> ann = type.getAnnotations();
                if (ann.size() > 1
                        || (ann.size() == 1 && !ann.contains(gutichecker.BOTTOM))) {
                    // the implicit default is BOTTOM, which cannot be used
                    // explicitly.
                    // if there are explicit annotations -> error.
                    reportError(type, p);
                }
            }
            return super.visitPrimitive(type, p);
        }

        /**
         * Each array dimension can only use at most one ownership modifier. The
         * check of the component type is done by the super method.
         */
        @Override
        public Void visitArray(AnnotatedArrayType type, Tree p) {
            if (type.getAnnotations().size() > 1) {
                // only one Universe modifier is allowed on the array
                reportError(type, p);
            }
            return super.visitArray(type, p);
        }
    }
}
